=== modified file 'bzrlib/selftest/testsftp.py'
--- bzrlib/selftest/testsftp.py	2005-10-27 03:28:48 +0000
+++ bzrlib/selftest/testsftp.py	2005-11-22 01:56:12 +0000
@@ -54,6 +54,8 @@
         threading.Thread.__init__(self)
         self._callback = callback
         self._socket = socket.socket()
+        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
+        self._socket.bind(('localhost', 0))
         self._socket.listen(1)
         self.port = self._socket.getsockname()[1]
         self.stop_event = threading.Event()
@@ -91,23 +93,105 @@
         TestCaseInTempDir.setUp(self)
         self._root = self.test_dir
 
+    def delayed_setup(self):
+        # some tests are just stubs that call setUp and then immediately call
+        # tearDwon.  so don't create the port listener until get_transport is
+        # called and we know we're in an actual test.
         self._listener = SingleListener(self._run_server)
         self._listener.setDaemon(True)
         self._listener.start()        
         self._sftp_url = 'sftp://foo:bar@localhost:%d/' % (self._listener.port,)
         
     def tearDown(self):
-        self._listener.stop()
+        try:
+            self._listener.stop()
+        except AttributeError:
+            pass
         TestCaseInTempDir.tearDown(self)
 
         
 class SFTPTransportTest (TestCaseWithSFTPServer, TestTransportMixIn):
     readonly = False
+    setup = True
 
     def get_transport(self):
+        if self.setup:
+            self.delayed_setup()
+            self.setup = False
         from bzrlib.transport.sftp import SFTPTransport
         url = self._sftp_url
         return SFTPTransport(url)
 
+    def test_sftp_locks(self):
+        from bzrlib.errors import LockError
+        t = self.get_transport()
+
+        l = t.lock_write('bogus')
+        self.failUnlessExists('bogus.write-lock')
+
+        # Don't wait for the lock, locking an already locked
+        # file should raise an assert
+        self.assertRaises(LockError, t.lock_write, 'bogus')
+
+        l.unlock()
+        self.failIf(os.path.lexists('bogus.write-lock'))
+
+        open('something.write-lock', 'wb').write('fake lock\n')
+        self.assertRaises(LockError, t.lock_write, 'something')
+        os.remove('something.write-lock')
+
+        l = t.lock_write('something')
+
+        l2 = t.lock_write('bogus')
+
+        l.unlock()
+        l2.unlock()
+
+
+class FakeSFTPTransport (object):
+    _sftp = object()
+fake = FakeSFTPTransport()
+
+
+class SFTPNonServerTest (unittest.TestCase):
+    def test_parse_url(self):
+        from bzrlib.transport.sftp import SFTPTransport
+        s = SFTPTransport('sftp://simple.example.com/%2fhome/source', clone_from=fake)
+        self.assertEquals(s._host, 'simple.example.com')
+        self.assertEquals(s._port, 22)
+        self.assertEquals(s._path, '/home/source')
+        self.assert_(s._password is None)
+        
+        s = SFTPTransport('sftp://ro%62ey:h%40t@example.com:2222/relative', clone_from=fake)
+        self.assertEquals(s._host, 'example.com')
+        self.assertEquals(s._port, 2222)
+        self.assertEquals(s._username, 'robey')
+        self.assertEquals(s._password, 'h@t')
+        self.assertEquals(s._path, 'relative')
+        
+
+class SFTPBranchTest(TestCaseWithSFTPServer):
+    """Test some stuff when accessing a bzr Branch over sftp"""
+
+    def test_lock_file(self):
+        """Make sure that a Branch accessed over sftp tries to lock itself."""
+        from bzrlib.branch import Branch
+
+        self.delayed_setup()
+        b = Branch.initialize(self._sftp_url)
+        self.failUnlessExists('.bzr/')
+        self.failUnlessExists('.bzr/branch-format')
+        self.failUnlessExists('.bzr/branch-lock')
+
+        self.failIf(os.path.lexists('.bzr/branch-lock.write-lock'))
+        b.lock_write()
+        self.failUnlessExists('.bzr/branch-lock.write-lock')
+        b.unlock()
+        self.failIf(os.path.lexists('.bzr/branch-lock.write-lock'))
+
+
 if not paramiko_loaded:
+    # TODO: Skip these
     del SFTPTransportTest
+    del SFTPNonServerTest
+    del SFTPBranchTest

=== modified file 'bzrlib/transport/sftp.py'
--- bzrlib/transport/sftp.py	2005-10-31 17:17:04 +0000
+++ bzrlib/transport/sftp.py	2005-11-22 01:56:12 +0000
@@ -23,10 +23,13 @@
 import stat
 import sys
 import urllib
+import time
+import random
 
 from bzrlib.errors import (FileExists, 
                            TransportNotPossible, NoSuchFile, NonRelativePath,
-                           TransportError)
+                           TransportError,
+                           LockError)
 from bzrlib.config import config_dir
 from bzrlib.trace import mutter, warning, error
 from bzrlib.transport import Transport, register_transport
@@ -36,6 +39,12 @@
 except ImportError:
     error('The SFTP transport requires paramiko.')
     raise
+else:
+    from paramiko.sftp import (SFTP_FLAG_WRITE, SFTP_FLAG_CREATE,
+                               SFTP_FLAG_EXCL, SFTP_FLAG_TRUNC,
+                               CMD_HANDLE, CMD_OPEN)
+    from paramiko.sftp_attr import SFTPAttributes
+    from paramiko.sftp_file import SFTPFile
 
 
 SYSTEM_HOSTKEYS = {}
@@ -81,13 +90,44 @@
 class SFTPTransportError (TransportError):
     pass
 
+class SFTPLock(object):
+    """This fakes a lock in a remote location."""
+    __slots__ = ['path', 'lock_path', 'lock_file', 'transport']
+    def __init__(self, path, transport):
+        assert isinstance(transport, SFTPTransport)
+
+        self.lock_file = None
+        self.path = path
+        self.lock_path = path + '.write-lock'
+        self.transport = transport
+        try:
+            self.lock_file = transport._sftp_open_exclusive(self.lock_path)
+        except FileExists:
+            raise LockError('File %r already locked' % (self.path,))
+
+    def __del__(self):
+        """Should this warn, or actually try to cleanup?"""
+        if self.lock_file:
+            warn("SFTPLock %r not explicitly unlocked" % (self.path,))
+            self.unlock()
+
+    def unlock(self):
+        if not self.lock_file:
+            return
+        self.lock_file.close()
+        self.lock_file = None
+        try:
+            self.transport.delete(self.lock_path)
+        except (NoSuchFile,):
+            # What specific errors should we catch here?
+            pass
 
 class SFTPTransport (Transport):
     """
     Transport implementation for SFTP access.
     """
 
-    _url_matcher = re.compile(r'^sftp://([^:@]*(:[^@]*)?@)?(.*?)(:\d+)?(/.*)?$')
+    _url_matcher = re.compile(r'^sftp://([^:@]*(:[^@]*)?@)?(.*?)(:[^/]+)?(/.*)?$')
     
     def __init__(self, base, clone_from=None):
         assert base.startswith('sftp://')
@@ -148,13 +188,14 @@
                 basepath.append(p)
 
         path = '/'.join(basepath)
-        if len(path) and path[0] != '/':
-            path = '/' + path
+        # could still be a "relative" path here, but relative on the sftp server
         return path
 
     def relpath(self, abspath):
         # FIXME: this is identical to HttpTransport -- share it
-        if not abspath.startswith(self.base):
+        m = self._url_matcher.match(abspath)
+        path = m.group(5)
+        if not path.startswith(self._path):
             raise NonRelativePath('path %r is not under base URL %r'
                            % (abspath, self.base))
         pl = len(self.base)
@@ -178,7 +219,13 @@
         """
         try:
             path = self._abspath(relpath)
-            return self._sftp.file(path)
+            f = self._sftp.file(path)
+            try:
+                f.prefetch()
+            except AttributeError:
+                # only works on paramiko 1.5.1 or greater
+                pass
+            return f
         except (IOError, paramiko.SSHException), x:
             raise NoSuchFile('Error retrieving %s: %s' % (path, str(x)), x)
 
@@ -194,8 +241,14 @@
                  Some implementations may return objects which can be read
                  past this length, but this is not guaranteed.
         """
+        # TODO: implement get_partial_multi to help with knit support
         f = self.get(relpath)
         f.seek(start)
+        try:
+            f.prefetch()
+        except AttributeError:
+            # only works on paramiko 1.5.1 or greater
+            pass
         return f
 
     def put(self, relpath, f):
@@ -205,18 +258,47 @@
         :param relpath: Location to put the contents, relative to base.
         :param f:       File-like or string object.
         """
-        # FIXME: should do something atomic or locking here, this is unsafe
-        try:
-            path = self._abspath(relpath)
-            fout = self._sftp.file(path, 'wb')
-        except IOError, e:
-            self._translate_io_exception(e, relpath)
-        except (IOError, paramiko.SSHException), x:
-            raise SFTPTransportError('Unable to write file %r' % (path,), x)
-        try:
-            self._pump(f, fout)
-        finally:
-            fout.close()
+        final_path = self._abspath(relpath)
+        tmp_relpath = '%s.tmp.%.9f.%d.%d' % (relpath, time.time(),
+                        os.getpid(), random.randint(0,0x7FFFFFFF))
+        tmp_abspath = self._abspath(tmp_relpath)
+        fout = self._sftp_open_exclusive(tmp_relpath)
+
+        try:
+            try:
+                self._pump(f, fout)
+            except IOError, e:
+                self._translate_io_exception(e, relpath)
+            except paramiko.SSHException, x:
+                raise SFTPTransportError('Unable to write file %r' % (path,), x)
+        except Exception, e:
+            # If we fail, try to clean up the temporary file
+            # before we throw the exception
+            # but don't let another exception mess things up
+            try:
+                fout.close()
+                self._sftp.remove(tmp_abspath)
+            except:
+                pass
+            raise e
+        else:
+            # sftp rename doesn't allow overwriting, so play tricks:
+            tmp_safety = 'bzr.tmp.%.9f.%d.%d' % (time.time(), os.getpid(), random.randint(0, 0x7FFFFFFF))
+            tmp_safety = self._abspath(tmp_safety)
+            try:
+                self._sftp.rename(final_path, tmp_safety)
+                file_existed = True
+            except:
+                file_existed = False
+            try:
+                self._sftp.rename(tmp_abspath, final_path)
+            except IOError, e:
+                self._translate_io_exception(e, relpath)
+            except paramiko.SSHException, x:
+                raise SFTPTransportError('Unable to rename into file %r' 
+                                          % (path,), x)
+            if file_existed:
+                self._sftp.unlink(tmp_safety)
 
     def iter_files_recursive(self):
         """Walk the relative paths of all files in this transport."""
@@ -326,7 +408,7 @@
     def lock_read(self, relpath):
         """
         Lock the given file for shared (read) access.
-        :return: A lock object, which should be passed to Transport.unlock()
+        :return: A lock object, which has an unlock() member function
         """
         # FIXME: there should be something clever i can do here...
         class BogusLock(object):
@@ -341,23 +423,25 @@
         Lock the given file for exclusive (write) access.
         WARNING: many transports do not support this, so trying avoid using it
 
-        :return: A lock object, which should be passed to Transport.unlock()
+        :return: A lock object, which has an unlock() member function
         """
-        # FIXME: there should be something clever i can do here...
-        class BogusLock(object):
-            def __init__(self, path):
-                self.path = path
-            def unlock(self):
-                pass
-        return BogusLock(relpath)
+        # This is a little bit bogus, but basically, we create a file
+        # which should not already exist, and if it does, we assume
+        # that there is a lock, and if it doesn't, the we assume
+        # that we have taken the lock.
+        return SFTPLock(relpath, self)
 
 
     def _unparse_url(self, path=None):
         if path is None:
             path = self._path
-        if self._port == 22:
-            return 'sftp://%s@%s%s' % (self._username, self._host, path)
-        return 'sftp://%s@%s:%d%s' % (self._username, self._host, self._port, path)
+        host = self._host
+        username = urllib.quote(self._username)
+        if self._password:
+            username += ':' + urllib.quote(self._password)
+        if self._port != 22:
+            host += ':%d' % self._port
+        return 'sftp://%s@%s/%s' % (username, host, urllib.quote(path))
 
     def _parse_url(self, url):
         assert url[:7] == 'sftp://'
@@ -368,16 +452,22 @@
         if self._username is None:
             self._username = getpass.getuser()
         else:
-            self._username = self._username[:-1]
-        if self._password:
-            self._password = self._password[1:]
-            self._username = self._username[len(self._password)+1:]
+            if self._password:
+                # username field is 'user:pass@' in this case, and password is ':pass'
+                username_len = len(self._username) - len(self._password) - 1
+                self._username = urllib.unquote(self._username[:username_len])
+                self._password = urllib.unquote(self._password[1:])
+            else:
+                self._username = urllib.unquote(self._username[:-1])
         if self._port is None:
             self._port = 22
         else:
             self._port = int(self._port[1:])
         if (self._path is None) or (self._path == ''):
-            self._path = '/'
+            self._path = ''
+        else:
+            # remove leading '/'
+            self._path = urllib.unquote(self._path[1:])
 
     def _sftp_connect(self):
         global SYSTEM_HOSTKEYS, BZR_HOSTKEYS
@@ -446,7 +536,10 @@
                 pass
 
         # give up and ask for a password
-        password = getpass.getpass('SSH %s@%s password: ' % (self._username, self._host))
+        # FIXME: shouldn't be implementing UI this deep into bzrlib
+        enc = sys.stdout.encoding
+        password = getpass.getpass('SSH %s@%s password: ' %
+            (self._username.encode(enc, 'replace'), self._host.encode(enc, 'replace')))
         try:
             transport.auth_password(self._username, password)
         except paramiko.SSHException:
@@ -460,7 +553,10 @@
             transport.auth_publickey(self._username, key)
             return True
         except paramiko.PasswordRequiredException:
-            password = getpass.getpass('SSH %s password: ' % (os.path.basename(filename),))
+            # FIXME: shouldn't be implementing UI this deep into bzrlib
+            enc = sys.stdout.encoding
+            password = getpass.getpass('SSH %s password: ' % 
+                (os.path.basename(filename).encode(enc, 'replace'),))
             try:
                 key = pkey_class.from_private_key_file(filename, password)
                 transport.auth_publickey(self._username, key)
@@ -472,3 +568,32 @@
         except IOError:
             pass
         return False
+
+    def _sftp_open_exclusive(self, relpath):
+        """Open a remote path exclusively.
+
+        SFTP supports O_EXCL (SFTP_FLAG_EXCL), which fails if
+        the file already exists. However it does not expose this
+        at the higher level of SFTPClient.open(), so we have to
+        sneak away with it.
+
+        WARNING: This breaks the SFTPClient abstraction, so it
+        could easily break against an updated version of paramiko.
+
+        :param relpath: The relative path, where the file should be opened
+        """
+        path = self._abspath(relpath)
+        attr = SFTPAttributes()
+        mode = (SFTP_FLAG_WRITE | SFTP_FLAG_CREATE 
+                | SFTP_FLAG_TRUNC | SFTP_FLAG_EXCL)
+        try:
+            t, msg = self._sftp._request(CMD_OPEN, path, mode, attr)
+            if t != CMD_HANDLE:
+                raise SFTPTransportError('Expected an SFTP handle')
+            handle = msg.get_string()
+            return SFTPFile(self._sftp, handle, 'w', -1)
+        except IOError, e:
+            self._translate_io_exception(e, relpath)
+        except paramiko.SSHException, x:
+            raise SFTPTransportError('Unable to open file %r' % (path,), x)
+

